"""
Classes that describe collections of droplets, i.e. emulsions, and their
temporal dynamics.


.. autosummary::
   :nosignatures:

   Emulsion
   EmulsionTimeCourse


.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import json
from typing import List  # @UnusedImport
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Generator,
    Iterator,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import numpy as np

from pde.fields import ScalarField
from pde.grids.base import GridBase
from pde.grids.cartesian import CartesianGridBase
from pde.storage.base import StorageBase
from pde.tools.cuboid import Cuboid
from pde.tools.misc import display_progress
from pde.tools.plotting import plot_on_axes
from pde.trackers.base import InfoDict
from pde.trackers.intervals import IntervalType

from .droplets import SphericalDroplet, droplet_from_data

if TYPE_CHECKING:
    from .trackers import DropletTracker  # @UnusedImport


DropletSequence = Union[Generator, Sequence[SphericalDroplet]]


class Emulsion(list):
    """ class representing a collection of droplets in a common system """

    def __init__(
        self,
        droplets: DropletSequence = None,
        grid: Optional[GridBase] = None,
        copy: bool = True,
    ):
        """
        Args:
            droplets:
                A list or generator of instances of
                :class:`~droplets.droplets.SphericalDroplet`.
            grid (:class:`~pde.grids.base.GridBase`):
                The grid on which the droplets are defined. This information can
                helpful to measure distances between droplets correctly.
            copy (bool, optional):
                Whether to make a copy of the droplet or not
        """
        # obtain grid of the emulsion
        if isinstance(droplets, Emulsion):
            self.grid: Optional[GridBase] = droplets.grid
            if grid is not None and self.grid != grid:
                raise ValueError("Emulsion grid is unequal to given grid")
        else:
            self.grid = grid

        # determine space dimension
        if grid is not None:
            self.dim: Optional[int] = grid.dim  # dimension of the space
        else:
            self.dim = None

        # add all droplets
        super().__init__()
        if droplets is not None:
            self.extend(droplets, copy=copy)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return super().__eq__(other) and self.grid == other.grid

    def __ne__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return super().__ne__(other) or self.grid != other.grid

    def __repr__(self):
        return f"{self.__class__.__name__}({super().__repr__()}, grid={self.grid})"

    def __add__(self, rhs):
        # return sum of Emulsions as an emulsion
        return Emulsion(list.__add__(self, rhs))

    def __getitem__(self, key: Union[int, slice]):
        # return result from extended slicing as Emulsion
        result = list.__getitem__(self, key)
        try:
            return Emulsion(result, grid=self.grid)
        except TypeError:
            return result

    def copy(self, min_radius: float = -1) -> "Emulsion":
        """return a copy of this emulsion

        Args:
            min_radius (float):
                The minimal radius of the droplets that are retained. Droplets
                with exactly min_radius are removed, so `min_radius == 0` can be
                used to filter vanished droplets.
        """
        droplets = [droplet.copy() for droplet in self if droplet.radius > min_radius]
        return self.__class__(
            droplets,
            grid=self.grid,
            copy=False,
        )

    def extend(  # type: ignore
        self,
        droplets: DropletSequence,
        copy: bool = True,
    ):
        """add many droplets to the emulsion

        Args:
            droplets (list): List of droplets to add to the emulsion
            copy (bool, optional): Whether to make a copy of the droplets or not
        """
        for droplet in droplets:
            self.append(droplet, copy=copy)

    def append(self, droplet: SphericalDroplet, copy: bool = True):
        """add a droplet to the emulsion

        Args:
            droplet (SphericalDroplet): Droplet to add to the emulsion
            copy (bool, optional): Whether to make a copy of the droplet or not
        """
        if self.dim is None:
            self.dim = droplet.dim
        elif self.dim != droplet.dim:
            raise ValueError(
                "Cannot append droplet in dimension "
                f"{droplet.dim} to emulsion in dimension "
                f"{self.dim}"
            )
        if copy:
            droplet = droplet.copy()
        super().append(droplet)

    @property
    def data(self):
        """numpy.ndarray: an array containing the data of the full emulsion

        This requires all droplets to be of the same class
        """
        if len(self) == 0:
            return None
        else:
            # emulsion contains at least one droplet
            classes = set(d.__class__ for d in self)
            if len(classes) > 1:
                raise TypeError(
                    "Emulsion data cannot be stored contiguously if it contains a "
                    f"multiple of droplet classes: "
                    + ", ".join(c.__name__ for c in classes)
                )
            return np.array([d.data for d in self])

    def get_linked_data(self) -> np.ndarray:
        """link the data of all droplets in a single array

        Returns:
            numpy.ndarray: The array containing all droplet data. If entries in
                this array are modified, it will be reflected in the droplets.
        """
        if len(self) == 0:
            data = np.empty(0)
        else:
            data = self.data  # create an array with all the droplet data
            # link back to droplets
            for i, d in enumerate(self):
                d.data = data[i]
        return data

    @classmethod
    def _from_hdf_dataset(cls, dataset, grid: Optional[GridBase] = None) -> "Emulsion":
        """construct an emulsion by reading data from an hdf5 dataset

        Args:
            dataset: an HDF5 dataset from which the data of the emulsion is read
            grid (GridBase): The grid on which the droplets are defined. This
                information is required to measure distances between droplets.
        """
        # there are values, so the emulsion is not empty
        droplet_class = dataset.attrs["droplet_class"]
        if droplet_class == "None":
            droplets: List[SphericalDroplet] = []
        else:
            droplets = [
                droplet_from_data(droplet_class, data)  # type: ignore
                for data in dataset
            ]
        return cls(droplets, grid=grid, copy=False)

    @classmethod
    def from_file(cls, filename: str) -> "Emulsion":
        """create emulsion by reading file

        Args:
            filename (str): Name of the file to read emulsion from
        """
        import h5py

        with h5py.File(filename, "r") as fp:
            if len(fp) != 1:
                raise RuntimeError(f"Multiple emulsions found in file `{filename}`")

            # read grid
            if "grid" in fp.attrs:
                grid: Optional[GridBase] = GridBase.from_state(fp.attrs["grid"])
            else:
                grid = None

            # read the actual droplet data
            dataset = fp[list(fp.keys())[0]]  # retrieve the only dataset
            obj = cls._from_hdf_dataset(dataset, grid)

        return obj

    def _write_hdf_dataset(self, hdf_path, key: str = "emulsion"):
        """ write data to a given hdf5 path `hdf_path` """
        if self:
            # emulsion contains at least one droplet
            dataset = hdf_path.create_dataset(key, data=self.data)
            # self.data ensures that there is only one droplet class
            dataset.attrs["droplet_class"] = self[0].__class__.__name__

        else:
            # create empty dataset to indicate empty emulsion
            dataset = hdf_path.create_dataset(key, shape=tuple())
            dataset.attrs["droplet_class"] = "None"

        return dataset

    def to_file(self, filename: str):
        """store data in hdf5 file

        Args:
            filename (str): Name of the file to write emulsion to
        """
        import h5py

        with h5py.File(filename, "w") as fp:
            # write the grid data
            if self.grid is not None:
                fp.attrs["grid"] = self.grid.state_serialized
            # write actual droplet data
            self._write_hdf_dataset(fp)

    @property
    def interface_width(self) -> Optional[float]:
        """float: the average interface width across all droplets

        This averages the interface widths of the individual droplets weighted
        by their surface area, i.e., the amount of interface
        """

        width, area = 0.0, 0.0
        for droplet in self:
            try:
                interface_width = droplet.interface_width
            except AttributeError:
                pass
            else:
                if interface_width is not None:
                    a = droplet.surface_area
                    width += interface_width * a
                    area += a

        if area == 0:
            return None
        else:
            return width / area

    @property
    def bbox(self) -> Cuboid:
        """ Cuboid: bounding box of the emulsion """
        if len(self) == 0:
            raise RuntimeError("Bounding box of empty emulsion is undefined")
        return sum((droplet.bbox for droplet in self[1:]), self[0].bbox)

    def get_phasefield(
        self, grid: GridBase = None, label: Optional[str] = None
    ) -> ScalarField:
        """create a phase field representing a list of droplets

        Args:
            grid (:class:`pde.grids.base.GridBase`):
                The grid on which the phase field is created. If omitted, the
                grid associated with the emulsion is used.
            label (str):
                Optional label for the returned scalar field

        Returns:
            :class:`pde.fields.scalar.ScalarField`: the actual phase field
        """
        if grid is None:
            grid = self.grid
        if grid is None:
            raise RuntimeError("Grid needs to be specified")

        if len(self) == 0:
            return ScalarField(grid)

        else:
            result: ScalarField = self[0].get_phase_field(grid, label=label)
            for d in self[1:]:
                result += d.get_phase_field(grid)  # type: ignore
            np.clip(result.data, 0, 1, out=result.data)
            return result

    def remove_small(self, min_radius: float = -np.inf):
        """remove droplets that are very small

        The emulsions is modified in-place.

        Args:
            min_radius (float): The minimal radius of the droplets that are
                retained. Droplets with exactly min_radius are removed, so
                `min_radius == 0` can be used to filter vanished droplets. The
                default value does not remove any droplets
        """
        for i in reversed(range(len(self))):
            if self[i].radius <= min_radius:
                self.pop(i)

    def get_pairwise_distances(self, subtract_radius: bool = False):
        """return the pairwise distance between droplets

        Args:
            subtract_radius (bool): determines whether the distance is measured
                from interface to interface (for round droplets) or center to
                center.

        Returns:
            np.array: a matrix with the distances between all droplets
        """
        if self.grid is None:

            def get_distance(p1, p2):
                """ helper function calculating the distance between points """
                return np.linalg.norm(p1 - p2)

        else:
            get_distance = self.grid.distance_real

        # calculate pairwise distance and return it in requested form
        num = len(self)
        dists = np.zeros((num, num))
        # iterate over all droplet pairs
        for i in range(num):
            for j in range(i + 1, num):
                d1, d2 = self[i], self[j]
                dist = get_distance(d1.position, d2.position)
                if subtract_radius:
                    dist -= d1.radius + d2.radius
                dists[i, j] = dists[j, i] = dist

        return dists

    def get_neighbor_distances(self, subtract_radius: bool = False):
        """calculates the distance of each droplet to its nearest neighbor

        Warning: Nearest neighbors are defined by comparing the distances
        between the centers of the droplets, not their surfaces.

        Args:
            subtract_radius (bool): Determines whether to subtract the radius
                from the distance, i.e., whether to return the distance between
                the surfaces instead of the positions

        Returns:
            np.array: a vector with a distance for each droplet
        """
        # handle simple cases
        if len(self) == 0:
            return np.zeros((0,))
        elif len(self) == 1:
            return np.full(1, np.nan)

        try:
            from scipy.spatial import cKDTree as KDTree
        except ImportError:
            from scipy.spatial import KDTree

        # build tree to query the nearest neighbors
        positions = self.data["position"]
        tree = KDTree(positions)
        dist, index = tree.query(positions, 2)

        if subtract_radius:
            return dist[:, 1] - self.data["radius"][index].sum(axis=1)
        else:
            return dist[:, 1]

    def remove_overlapping(self, min_distance: float = 0):
        """remove all droplets that are overlapping.

        If a pair of overlapping droplets was found, the smaller one of these
        is removed from the current emulsion. This method modifies the emulsion
        in place and thus does not return anything.

        Args:
            min_distance (float): The minimal distance droplets need to be
                apart. The default value of 0 corresponds to just remove
                overlapping droplets. Larger values ensure that droplets keep a
                distance, while negative values allow for some overlap.
        """
        # filter duplicates until there are none left
        dists = self.get_pairwise_distances(subtract_radius=True)
        np.fill_diagonal(dists, np.inf)

        while len(dists) > 1:
            # find minimal distance
            x, y = np.unravel_index(np.argmin(dists), dists.shape)
            if dists[x, y] < min_distance:
                # droplets overlap -> remove the smaller one
                if self[x].radius > self[y].radius:
                    self.pop(y)
                    dists = np.delete(np.delete(dists, y, 0), y, 1)
                else:
                    self.pop(x)
                    dists = np.delete(np.delete(dists, x, 0), x, 1)
            else:
                break

    @property
    def total_droplet_volume(self) -> float:
        """ float: the total volume of all droplets """
        return sum(droplet.volume for droplet in self)

    def get_size_statistics(self) -> Dict[str, float]:
        """determine size statistics of the current emulsion

        Returns:
            dict: a dictionary with various size statistics
        """
        if len(self) == 0:
            return {
                "count": 0,
                "radius_mean": np.nan,
                "radius_std": np.nan,
                "volume_mean": np.nan,
                "volume_std": np.nan,
            }

        radii = [droplet.radius for droplet in self]
        volumes = [droplet.volume for droplet in self]
        return {
            "count": len(self),
            "radius_mean": np.mean(radii),
            "radius_std": np.std(radii),
            "volume_mean": np.mean(volumes),
            "volume_std": np.std(volumes),
        }

    @plot_on_axes()
    def plot(
        self,
        ax,
        field: ScalarField = None,
        image_args: Dict[str, Any] = None,
        repeat_periodically: bool = True,
        **kwargs,
    ):
        """plot the current emulsion together with a corresponding field

        Args:
            ax (:class:`matplotlib.axes.Axes`):
                The axes in which the background is shown
            field (:class:`pde.fields.scalar.ScalarField`):
                provides the phase field that is shown as a background
            image_args (dict):
                additional arguments determining how the phase field in the
                background is plotted. Acceptable arguments are described in
                :func:`~pde.fields.base.FieldBase.plot`.
            repeat_periodically (bool):
                flag determining whether droplets are shown on both sides of
                periodic boundary conditions. This option can slow down plotting
            **kwargs:
                Additional keyword arguments are passed to the function creating the
                patch that represents the droplet. For instance, to only draw the
                outlines of the droplets, you may need to supply `fill=False`.
        """
        if self.dim != 2:
            raise NotImplementedError(
                f"Plotting emulsions in {self.dim} dimensions is not implemented."
            )
        grid_compatible = (
            self.grid is None or field is None or self.grid.compatible_with(field.grid)
        )
        if not grid_compatible:
            raise ValueError("Emulsion grid is not compatible with field grid")
        grid = self.grid

        # plot background and determine bounds for the droplets
        if field is not None:
            # plot the phase field and use its bounds
            if image_args is None:
                image_args = {}
            field.plot(kind="image", ax=ax, **image_args)
            ax.autoscale(False)  # fix image bounds to phase field
            if grid is None:
                grid = field.grid

        else:
            if isinstance(grid, CartesianGridBase):
                # determine the bounds from the (2d) grid
                bounds = grid.axes_bounds
            else:
                # determine the bounds from the emulsion data itself
                bounds = self.bbox.bounds
            ax.set_xlim(*bounds[0])
            ax.set_ylim(*bounds[1])
            ax.set_aspect("equal")

        # get patches representing all droplets
        if grid is None or not repeat_periodically:
            # plot only the droplets themselves
            patches = [droplet._get_mpl_patch(**kwargs) for droplet in self]
        else:
            # plot droplets also in their mirror positions
            patches = []
            for droplet in self:
                for p in grid.iter_mirror_points(
                    droplet.position, with_self=True, only_periodic=True
                ):
                    # create copy with changed position
                    d = droplet.copy(position=p)
                    patches.append(d._get_mpl_patch(**kwargs))

        # add all patches as a collection
        import matplotlib as mpl

        coll = mpl.collections.PatchCollection(patches, match_original=True)
        ax.add_collection(coll)


class EmulsionTimeCourse:
    """ represents emulsions as a function of time """

    def __init__(self, emulsions=None, times=None):
        """
        Args:
            emulsions (list): List of emulsions that describe this time course
            times (list): Times associated with the emulsions
        """
        if isinstance(emulsions, EmulsionTimeCourse):
            # extract data from given object; ignore `times`
            times = emulsions.times
            self.grid = emulsions.grid
            emulsions = emulsions.emulsions
        else:
            self.grid = None

        self.emulsions = []
        self.times = []

        # add all emulsions
        if emulsions:
            for e in emulsions:
                self.append(Emulsion(e))

        # add all times
        if times is not None:
            self.times = list(times)

        if len(self.times) != len(self.emulsions):
            raise ValueError(
                "The list of emulsions and the list of times need "
                "to have the same length"
            )

    def append(
        self, emulsion: Emulsion, time: Optional[float] = None, copy: bool = True
    ):
        """add an emulsion to the list

        Args:
            emulsions (Emulsion): An :class:`Emulsion` instance that is added
                to the time course
            time (float): The time point associated with this emulsion
            copy (bool): Whether to copy the emulsion
        """
        emulsion = Emulsion(emulsion)  # make sure this is an emulsion

        # check grid consistency
        if self.grid is None:
            self.grid = emulsion.grid
        elif emulsion.grid and not self.grid.compatible_with(emulsion.grid):
            raise ValueError(
                "Grid of the EmulsionTimeCourse is not compatible with the grid of the "
                f"emulsion to be added. ({self.grid} != {emulsion.grid})"
            )
        # add the emulsion
        if copy:
            emulsion = emulsion.copy()
        self.emulsions.append(emulsion)

        if time is None:
            time = 0 if len(self.times) == 0 else self.times[-1] + 1
        self.times.append(time)

    def clear(self):
        """ removes all data stored in this instance """
        self.emulsions = []
        self.times = []

    def __repr__(self):
        return (
            f"{self.__class__.__name__}(emulsions=Array({len(self)}), "
            f"times=Array({len(self)}), grid={self.grid})"
        )

    def __len__(self):
        return len(self.times)

    def __getitem__(self, key: Union[int, slice]):
        """ return the information for the given index """
        result = self.emulsions.__getitem__(key)
        if isinstance(key, slice):
            return self.__class__(emulsions=result, times=self.times[key])
        else:
            return result

    def __iter__(self) -> Iterator[Tuple[float, Emulsion]]:
        """ iterate over the emulsions """
        return iter(self.emulsions)

    def items(self):
        """ iterate over all times and emulsions, returning them in pairs """
        return zip(self.times, self.emulsions)

    def __eq__(self, other):
        """ determine whether two EmulsionTimeCourse instance are equal """
        grids_equal = self.grid is None or other.grid is None or self.grid == other.grid
        return (
            grids_equal
            and self.times == other.times
            and self.emulsions == other.emulsions
        )

    @classmethod
    def from_storage(
        cls, storage: StorageBase, progress: bool = True, **kwargs
    ) -> "EmulsionTimeCourse":
        r"""create an emulsion time course from a stored phase field

        Args:
            storage (:class:`~pde.storage.base.StorageBase`):
                The phase fields for many time instances
            progress (bool):
                Whether to show the progress of the process
            \**kwargs:
                All other parameters are forwarded to the
                :meth:`~droplets.image_analysis.locate_droplets`.

        Returns:
            EmulsionTimeCourse: an instance describing the emulsion time course
        """
        from .image_analysis import locate_droplets

        result = cls()

        it = display_progress(storage.items(), total=len(storage), enabled=progress)
        for t, frame in it:
            emulsion = locate_droplets(frame, **kwargs)
            result.append(emulsion, time=t)
        return result

    @classmethod
    def from_file(cls, filename: str, progress: bool = True) -> "EmulsionTimeCourse":
        """create emulsion time course by reading file

        Args:
            filename (str): The filename from which the emulsion is read
            progress (bool): Whether to show the progress of the process

        Returns:
            EmulsionTimeCourse: an instance describing the emulsion time course
        """
        import h5py

        obj = cls()
        with h5py.File(filename, "r") as fp:
            # read grid
            if "grid" in fp.attrs:
                grid: Optional[GridBase] = GridBase.from_state(fp.attrs["grid"])
            else:
                grid = None

            # load the actual emulsion data and iterate in the right order
            for key in display_progress(
                sorted(fp.keys()), total=len(fp), enabled=progress
            ):
                dataset = fp[key]
                obj.append(
                    Emulsion._from_hdf_dataset(dataset, grid),
                    time=dataset.attrs["time"],
                )
        return obj

    def to_file(self, filename: str, info: InfoDict = None):
        """store data in hdf5 file

        Args:
            filename (str): determines the location where the file is written
            info (dict): can be additional data stored alongside
        """
        import h5py

        with h5py.File(filename, "w") as fp:
            # write the actual emulsion data
            for i, (time, emulsion) in enumerate(self.items()):
                dataset = emulsion._write_hdf_dataset(fp, f"time_{i:06d}")
                dataset.attrs["time"] = time

            # write additional information
            if info:
                for k, v in info.items():
                    fp.attrs[k] = json.dumps(v)

            # write the grid data -> this might overwrite grid data that is
            # present in the info dictionary, but in normal cases these grids
            # should be identical, so we don't handle this case explicitly
            if self.grid is not None:
                fp.attrs["grid"] = self.grid.state_serialized

    def tracker(
        self,
        interval: Union[int, float, IntervalType] = 1,
        filename: Optional[str] = None,
    ) -> "DropletTracker":
        """return a tracker that analyzes emulsions during simulations

        Args:
            interval: Determines how often the tracker interrupts the
                simulation. Simple numbers are interpreted as durations measured
                in the simulation time variable. Alternatively, instances of
                :class:`~droplets.simulation.trackers.LogarithmicIntervals` and
                :class:`~droplets.simulation.trackers.RealtimeIntervals` might
                be given for more control.
            filename (str): determines where the EmulsionTimeCourse data is
                stored
        """
        from .trackers import DropletTracker  # @Reimport

        return DropletTracker(
            emulsion_timecourse=self, filename=filename, interval=interval
        )


__all__ = ["Emulsion", "EmulsionTimeCourse"]



"""
Classes representing the time evolution of droplets

.. autosummary::
   :nosignatures:

   DropletTrack
   DropletTrackList


.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import json
from typing import TYPE_CHECKING, Optional, Union

import numpy as np
from numpy.lib import recfunctions as rfn
from scipy.ndimage import filters

from pde.grids.base import GridBase
from pde.trackers.base import InfoDict

from .droplets import SphericalDroplet, droplet_from_data

if TYPE_CHECKING:
    from .emulsions import EmulsionTimeCourse  # @UnusedImport


def contiguous_true_regions(condition: np.ndarray) -> np.ndarray:
    """Finds contiguous True regions in the boolean array "condition"

    Inspired by http://stackoverflow.com/a/4495197/932593

    Args:
        condition (:class:`numpy.ndarray`):
            A one-dimensional boolean array


    Returns:
        :class:`numpy.ndarray`: A two-dimensional array where the first column
        is the start index of the region and the second column is the end index
    """
    if len(condition) == 0:
        return []

    # convert condition array to integer
    condition = np.asarray(condition, np.int)

    # Find the indices of changes in "condition"
    d = np.diff(condition)
    idx = np.flatnonzero(d)

    # We need to start things after the change in "condition". Therefore,
    # we'll shift the index by 1 to the right.
    idx += 1

    if condition[0]:
        # If the start of condition is True prepend a 0
        idx = np.r_[0, idx]

    if condition[-1]:
        # If the end of condition is True, append the length of the array
        idx = np.r_[idx, condition.size]

    # Reshape the result into two columns
    return idx.reshape(-1, 2)


class DropletTrack:
    """ information about a single droplet over multiple time steps """

    def __init__(self, droplets=None, times=None):
        """
        Args:
            emulsions (list): List of emulsions that describe this time course
            times (list): Times associated with the emulsions
        """
        if isinstance(droplets, DropletTrack):
            # get data from given object
            times = droplets.times
            droplets = droplets.droplets

        self.droplets = []
        self.times = []

        # add all emulsions
        if droplets:
            for d in droplets:
                self.append(d)

        # add all times
        if times is not None:
            self.times = list(times)

        if len(self.times) != len(self.droplets):
            raise ValueError(
                "The lists of droplets and times need to have the same length"
            )

    def __repr__(self):
        """ human-readable representation of a droplet track """
        class_name = self.__class__.__name__
        if len(self.times) == 0:
            return f"{class_name}([])"
        elif len(self.times) == 1:
            return f"{class_name}(time={self.start})"
        else:
            return f"{class_name}(timespan={self.start}..{self.end})"

    def __len__(self):
        """ number of time points """
        return len(self.times)

    def __getitem__(self, key: Union[int, slice]):
        """ return the droplets identified by the given index/slice """
        result = self.droplets.__getitem__(key)
        if isinstance(key, slice):
            return self.__class__(droplets=result, times=self.times[key])
        else:
            return result

    def __eq__(self, other):
        """ determine whether two DropletTracks instance are equal """
        return self.times == other.times and self.droplets == other.droplets

    @property
    def start(self) -> float:
        """ float: first time point """
        return self.times[0]  # type: ignore

    @property
    def end(self) -> float:
        """ float: last time point """
        return self.times[-1]  # type: ignore

    @property
    def duration(self) -> float:
        """ float: total duration of the track """
        if len(self.times) > 0:
            return self.end - self.start
        else:
            return 0

    @property
    def first(self) -> SphericalDroplet:
        """ SphericalDroplet: first droplet instance """
        return self.droplets[0]  # type: ignore

    @property
    def last(self) -> SphericalDroplet:
        """ SphericalDroplet: last droplet instance """
        return self.droplets[-1]  # type: ignore

    @property
    def dim(self) -> Optional[int]:
        """ return the space dimension of the droplets """
        try:
            return self.last.dim
        except IndexError:
            return None

    @property
    def data(self):
        """ numpy.ndarray: an array containing the data of the full track """
        if len(self) == 0:
            return None
        else:
            d0 = self.first
            dtype = [("time", "f8")] + d0.data.dtype.descr
            result = np.empty(len(self), dtype=dtype)
            for i in range(len(self)):
                result[i] = (self.times[i],) + self.droplets[i].data.tolist()
            return result

    def __iter__(self):
        """ iterate over all droplets """
        return iter(self.droplets)

    def items(self):
        """ iterate over all times and droplets, returning them in pairs """
        return zip(self.times, self.droplets)

    def append(self, droplet: SphericalDroplet, time: Optional[float] = None):
        """append a new droplet with a time code

        Args:
            droplet (SphericalDroplet): the droplet
            time (float, optional): The time point
        """
        if self.dim is not None and droplet.dim != self.dim:
            raise ValueError(
                "Space dimension of added droplet must match the DropletTrack "
                f" ({droplet.dim} != {self.dim})"
            )

        # add the emulsion
        self.droplets.append(droplet.copy())
        if time is None:
            time = 0 if len(self.times) == 0 else self.times[-1] + 1
        self.times.append(time)

    def get_position(self, time: float):
        """ numpy.ndarray: returns the droplet position at a specific time """
        try:
            idx = self.times.index(time)
        except AttributeError:
            # assume that self.times is a numpy array
            idx = np.nonzero(self.times == time)[0][0]
        return self.droplets[idx].position

    def get_trajectory(self, smoothing: float = 0):
        """return a list of positions over time

        Args:
            smoothing (float): Determines the length scale for some gaussian
                smoothing of the trajectory. Setting this to zero disables
                smoothing.

        Returns:
            numpy.ndarray: An array giving the position of the droplet at each
                time instance
        """
        trajectory = np.array([droplet.position for droplet in self.droplets])
        if smoothing:
            filters.gaussian_filter1d(
                trajectory, output=trajectory, sigma=smoothing, axis=0, mode="nearest"
            )
        return trajectory

    def get_radii(self) -> np.ndarray:
        """ returns the droplet radius for each time point """
        return np.array([droplet.radius for droplet in self.droplets])

    def get_volumes(self) -> np.ndarray:
        """ returns the droplet volume for each time point """
        return np.array([droplet.volume for droplet in self.droplets])

    def time_overlaps(self, other: "DropletTrack") -> bool:
        """determine whether two DropletTrack instances overlaps in time

        Args:
            other (DropletTrack): The other droplet track

        Returns:
            bool: True when both tracks contain droplets at the same time step
        """
        s0, s1 = self.start, self.end
        o0, o1 = other.start, other.end
        return s0 <= o1 and o0 <= s1

    @classmethod
    def _from_hdf_dataset(cls, dataset) -> "DropletTrack":
        """construct a droplet track by reading data from an hdf5 dataset

        Args:
            dataset:
                an HDF5 dataset from which the data of the droplet track is read
        """
        # there are values, so the emulsion is not empty
        droplet_class = dataset.attrs["droplet_class"]
        obj = cls()
        if droplet_class == "None":
            return obj
        else:
            # separate time from the data set
            times = dataset["time"]
            droplet_data = rfn.rec_drop_fields(dataset, "time")
            for time, data in zip(times, droplet_data):
                droplet = droplet_from_data(droplet_class, data)
                obj.append(droplet, time=time)  # type: ignore

        return obj

    @classmethod
    def from_file(cls, filename: str) -> "DropletTrack":
        """create droplet track by reading from file

        Args:
            filename (str): Name of the file to read emulsion from
        """
        import h5py

        with h5py.File(filename, "r") as fp:
            if len(fp) != 1:
                raise RuntimeError(f"Multiple droplet tracks found in file {filename}")
            dataset = fp[list(fp.keys())[0]]  # retrieve the only dataset
            obj = cls._from_hdf_dataset(dataset)

        return obj

    def _write_hdf_dataset(self, hdf_path, key: str = "droplet_track"):
        """ write data to a given hdf5 path `hdf_path` """
        if self:
            # emulsion contains at least one droplet
            dataset = hdf_path.create_dataset(key, data=self.data)
            # self.data ensures that there is only one droplet class
            dataset.attrs["droplet_class"] = self[0].__class__.__name__

        else:
            # create empty dataset to indicate empty emulsion
            dataset = hdf_path.create_dataset(key, shape=tuple())
            dataset.attrs["droplet_class"] = "None"

        return dataset

    def to_file(self, filename: str, info: InfoDict = None):
        """store data in hdf5 file

        Args:
            filename (str): Name of the file to write emulsion to
        """
        import h5py

        with h5py.File(filename, "w") as fp:
            self._write_hdf_dataset(fp)

            # write additional information
            if info:
                for k, v in info.items():
                    fp.attrs[k] = json.dumps(v)

    def plot(self, attribute: str = "radius", **kwargs):
        """plot the time evolution of the droplet

        Args:
            attribute (str):
                The attribute to plot. Typical values include `radius` and
                `volume`, but others might be defined on the droplet class.
            **kwargs:
                Additional keyword arguments are passed to the matplotlib plot
                function to affect the appearance.
        """
        import matplotlib.pyplot as plt

        if len(self.times) == 0:
            return

        if attribute in {"radius", "radii"}:
            data = self.get_radii()
            ylabel = "Radius"
        elif attribute in {"volume", "volumes"}:
            data = self.get_volumes()
            ylabel = "Volume"
        else:
            data = [getattr(droplet, attribute) for droplet in self.droplets]
            ylabel = attribute.capitalize()

        plt.plot(self.times, data, **kwargs)
        plt.xlabel("Time")
        plt.ylabel(ylabel)

    def plot_positions(
        self, grid: Optional[GridBase] = None, arrow: bool = True, **kwargs
    ):
        """plot the droplet track

        Args:
            grid (GridBase, optional): The grid on which the droplets are
                defined. If given, periodic boundary conditions can be respected
                in the plotting.
            arrow (bool, optional): Flag determining whether an arrow head is
                shown to indicate the direction of the droplet drift.
            **kwargs:
                Additional keyword arguments are passed to the matplotlib plot
                function to affect the appearance.
        """
        import matplotlib.pyplot as plt

        if len(self.times) == 0:
            return

        if self.dim != 2:
            raise NotImplementedError(
                "Plotting only implemented for two-dimensional grids"
            )

        # obtain droplet positions as a function of time
        xy = self.get_trajectory()

        if grid is None:
            # simply plot the trajectory
            (line,) = plt.plot(xy[:, 0], xy[:, 1], **kwargs)

        else:
            # use the grid to detect wrapping around
            segments = []
            for p1, p2 in zip(xy[:-1], xy[1:]):
                dist_direct = np.hypot(p1[0] - p2[0], p1[1] - p2[1])
                dist_real = grid.distance_real(p1, p2)
                close = np.isclose(dist_direct, dist_real)
                segments.append(close)

            # plot the individual segments
            line = None
            for s, e in contiguous_true_regions(segments):
                if line is None:
                    color = kwargs.get("color")
                else:
                    color = line.get_color()  # ensure colors stays the same
                (line,) = plt.plot(xy[s : e + 1, 0], xy[s : e + 1, 1], color=color)

        if arrow and len(xy) >= 2:
            size = min(sum(plt.xlim()), sum(plt.ylim()))
            dx = xy[-1] - xy[-2]
            plt.arrow(
                xy[-2, 0],
                xy[-2, 1],
                dx[0],
                dx[1],
                head_width=0.02 * size,
                color=line.get_color(),
            )


class DropletTrackList(list):
    """ a list of instances of :class:`DropletTrack` """

    @classmethod
    def from_emulsion_time_course(cls, time_course: "EmulsionTimeCourse"):
        """ create the instance from an emulsion time course """
        # get tracks, i.e. clearly overlapping droplets
        tracks = cls()
        t_last = None
        for t, emulsion in time_course.items():
            # determine live tracks
            tracks_alive = [track for track in tracks if track.end <= t_last]

            # handle all droplets in the emulsion
            for droplet in emulsion:
                overlap_count = 0
                overlap_track: Optional[DropletTrack] = None
                # iterate over all active tracks
                for track in tracks_alive:
                    if track.last.overlaps(droplet, time_course.grid):
                        overlap_count += 1
                        overlap_track = track

                if overlap_count == 1:
                    overlap_track.append(droplet, time=t)  # type: ignore
                else:
                    tracks.append(DropletTrack(droplets=[droplet], times=[t]))

            t_last = t
        return tracks

    def __getitem__(self, key: Union[int, slice]):
        """ return the droplets identified by the given index/slice """
        result = super().__getitem__(key)
        if isinstance(key, slice):
            return self.__class__(result)
        else:
            return result

    def remove_short_tracks(self, min_duration: float = 0):
        """remove tracks that a shorter than a minimal duration

        Args:
            min_duration (float): The minimal duration a droplet track must have
                in order to be retained. This is measured in actual time and not
                in the number of time steps stored in the track
        """
        for i in reversed(range(len(self))):
            if self[i].duration <= min_duration:
                self.pop(i)

    @classmethod
    def from_file(cls, filename: str) -> "DropletTrackList":
        """create droplet track list by reading file

        Args:
            filename (str): The filename from which the data is read

        Returns:
            DropletTrackList: an instance describing the droplet track list
        """
        import h5py

        obj = cls()
        with h5py.File(filename, "r") as fp:
            for key in sorted(fp.keys()):  # iterate in the stored order
                dataset = fp[key]
                obj.append(DropletTrack._from_hdf_dataset(dataset))
        return obj

    def to_file(self, filename: str, info: InfoDict = None):
        """store data in hdf5 file

        Args:
            filename (str): determines the location where the file is written
            info (dict): can be additional data stored alongside
        """
        import h5py

        with h5py.File(filename, "w") as fp:
            # write the actual emulsion data
            for i, droplet_track in enumerate(self):
                droplet_track._write_hdf_dataset(fp, f"track_{i:06d}")

            # write additional information
            if info:
                for k, v in info.items():
                    fp.attrs[k] = json.dumps(v)

    def plot(self, attribute: str = "radius", **kwargs):
        """plot the time evolution of all droplets

        Args:
            attribute (str):
                The attribute to plot. Typical values include `radius` and
                `volume`, but others might be defined on the droplet class.
            **kwargs:
                Additional keyword arguments are passed to the matplotlib plot
                function to affect the appearance.
        """
        kwargs.setdefault("color", "k")
        # adjust alpha such that multiple tracks are visible well
        kwargs.setdefault("alpha", min(0.8, 20 / len(self)))
        for track in self:
            track.plot(attribute=attribute, **kwargs)

    def plot_positions(self, **kwargs):
        """plot all droplet tracks

        Args:
            **kwargs:
                Additional keyword arguments are passed to the matplotlib plot
                function to affect the appearance.
        """
        for track in self:
            track.plot_positions(**kwargs)


__all__ = ["DropletTrack", "DropletTrackList"]



"""
Classes representing (perturbed) droplets in various dimensions


.. autosummary::
   :nosignatures:

   SphericalDroplet
   DiffuseDroplet
   PerturbedDroplet2D
   PerturbedDroplet3D


Inheritance structure of the classes:

.. inheritance-diagram:: droplets.droplets
   :parts: 1

The details of the classes are explained below:

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import logging
from abc import ABCMeta, abstractmethod
from typing import Dict  # @UnusedImport
from typing import TYPE_CHECKING, List, Optional, Sequence, TypeVar

import numpy as np
from numpy.lib.recfunctions import structured_to_unstructured
from scipy import integrate

from pde.fields import ScalarField
from pde.grids.base import GridBase
from pde.tools import spherical
from pde.tools.cuboid import Cuboid
from pde.tools.misc import preserve_scalars
from pde.tools.plotting import plot_on_axes

# work-around to satisfy type checking in python 3.6
if TYPE_CHECKING:
    # TYPE_CHECKING will always be False in production and this circular import
    # will thus be resolved.
    from .emulsions import EmulsionTimeCourse  # @UnusedImport


TDroplet = TypeVar("TDroplet", bound="DropletBase")


def get_dtype_field_size(dtype, field_name: str) -> int:
    """return the number of elements in a field of structured numpy array

    Args:
        dtype (list):
            The dtype of the numpy array
        field_name (str):
            The name of the field that needs to be checked
    """
    shape = dtype.fields[field_name][0].shape
    return np.prod(shape) if shape else 1  # type: ignore


def iterate_in_pairs(it, fill=0):
    """return consecutive pairs from an iterator

    For instance, `list(pair_iterator('ABCDE', fill='Z'))` returns
    `[('A', 'B'), ('C', 'D'), ('E', 'Z')]`

    Args:
        it (iterator): The iterator
        fill: The value returned for the second part of the last returned tuple
            if the length of the iterator is odd

    Returns:
        This is a generator function that yields pairs of items of the iterator
    """
    it = iter(it)
    while True:
        # obtain first item of tuple
        try:
            first = next(it)
        except StopIteration:
            break
        # obtain second item of tuple
        try:
            yield first, next(it)
        except StopIteration:
            yield first, fill
            break


class DropletBase:
    """represents a generic droplet

    The data associated with a droplet is stored in structured array.
    Consequently, the `dtype` of the array determines what information the
    droplet class stores.
    """

    _subclasses: Dict[str, "DropletBase"] = {}  # collect all inheriting classes

    __slots__ = ["data"]

    @classmethod
    def from_data(cls, data: np.recarray) -> "DropletBase":
        """create droplet class from a given data

        Args:
            data (:class:`numpy.recarray`):
                The data array used to initialize the droplet
        """
        # note that we do not call the __init__ method of the class, since we do
        # not need to invent the dtype and set all the attributes. We here
        # simply assume that the given data is sane
        obj = cls.__new__(cls)
        obj.data = data
        return obj  # type: ignore

    @classmethod
    def from_droplet(cls, droplet: "DropletBase", **kwargs) -> "DropletBase":
        r"""return a droplet with data taken from `droplet`

        Args:
            droplet (:class:`DropletBase`):
                Another droplet from which data is copied. This does not be the
                same type of droplet
            \**kwargs:
                Additional arguments that can overwrite data in `droplet` or
                set additional arguments for the current class
        """
        args = droplet._args
        args.update(kwargs)
        return cls(**args)  # type: ignore

    def _init_data(self, **kwargs):
        """initializes the `data` attribute if it is not present

        Args:
            **kwargs:
                Arguments used to determine the dtype of the data array
        """
        if not hasattr(self, "data"):
            dtype = self.get_dtype(**kwargs)

            # We need to create an empty record with the correct data type. Note
            # that the conversion np.recarray(1)[0] turns the array into a
            # scalar type (instance of numpy.record) that contains the
            # structured data. This conversion is necessary for numba to operate
            # on the data.
            self.data = np.recarray(1, dtype=dtype)[0]

    def __init_subclass__(cls, **kwargs):  # @NoSelf
        """ register all subclassess to reconstruct them later """
        super().__init_subclass__(**kwargs)
        cls._subclasses[cls.__name__] = cls

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        return np.allclose(
            self._data_array, other._data_array, rtol=0, atol=0, equal_nan=True
        )

    def check_data(self):
        """ method that checks the validity and consistency of self.data """
        pass

    @property
    def _args(self):
        return {key: self.data[key] for key in self.data.dtype.names}

    def __str__(self):
        arg_list = [f"{key}={value}" for key, value in self._args.items()]
        return f"{self.__class__.__name__}({', '.join(arg_list)})"

    __repr__ = __str__

    @property
    def _data_array(self) -> np.ndarray:
        """ return the data of the droplet in an unstructured array """
        return structured_to_unstructured(self.data)

    def copy(self: TDroplet, **kwargs) -> TDroplet:
        r"""return a copy of the current droplet

        Args:
            \**kwargs: Additional arguments an be used to set data of the
                returned droplet.
        """
        if kwargs:
            return self.from_droplet(self, **kwargs)  # type: ignore
        else:
            return self.from_data(self.data.copy())  # type: ignore

    @property
    def data_bounds(self):
        """ lower and upper bounds on the parameters """
        num = len(self._data_array)
        return np.full(num, -np.inf), np.full(num, np.inf)


class SphericalDroplet(DropletBase):  # lgtm [py/missing-equals]
    """ Represents a single, spherical droplet """

    __slots__ = ["data"]

    def __init__(self, position: Sequence[float], radius: float):
        r"""
        Args:
            position (:class:`numpy.ndarray`):
                Position of the droplet center
            radius (float):
                Radius of the droplet
        """
        self._init_data(position=position)

        self.position = position
        self.radius = radius
        self.check_data()

    def check_data(self):
        """ method that checks the validity and consistency of self.data """
        if self.radius < 0:
            raise ValueError("Radius must be positive")

    @classmethod
    def get_dtype(cls, position):
        position = np.atleast_1d(position)
        dim = len(position)
        return [("position", float, (dim,)), ("radius", float)]

    @property
    def dim(self) -> int:
        return get_dtype_field_size(self.data.dtype, "position")

    @property
    def data_bounds(self):
        """ lower and upper bounds on the parameters """
        l, h = super().data_bounds
        l[self.dim] = 0  # radius must be non-negative
        return l, h

    @classmethod
    def from_volume(cls, position: Sequence[float], volume: float):
        """Construct a droplet from given volume instead of radius

        Args:
            position (vector): center of the droplet
            volume (float): volume of the droplet
            interface_width (float, optional): width of the interface
        """
        dim = len(np.array(position, np.double, ndmin=1))
        radius = spherical.radius_from_volume(volume, dim)
        return cls(position, radius)

    @property
    def position(self):
        return self.data["position"]

    @position.setter
    def position(self, value):
        value = np.asanyarray(value)
        if len(value) != self.dim:
            raise ValueError(f"Length of position must be {self.dim}")
        self.data["position"] = value

    @property
    def radius(self) -> float:
        return float(self.data["radius"])

    @radius.setter
    def radius(self, value: float):
        self.data["radius"] = value
        self.check_data()

    @property
    def volume(self) -> float:
        """ float: volume of the droplet """
        return spherical.volume_from_radius(self.radius, self.dim)

    @volume.setter
    def volume(self, volume: float):
        """ set the radius from a supplied volume """
        self.radius = spherical.radius_from_volume(volume, self.dim)

    @property
    def surface_area(self) -> float:
        """ float: surface area of the droplet """
        return spherical.surface_from_radius(self.radius, self.dim)

    @property
    def bbox(self) -> Cuboid:
        """ Cuboid: bounding box of the droplet """
        return Cuboid.from_points(
            self.position - self.radius, self.position + self.radius
        )

    def overlaps(self, other: "SphericalDroplet", grid: GridBase = None) -> bool:
        """determine whether another droplet overlaps with this one

        Note that this function so far only compares the distances of the
        droplets to their radii, which does not respect perturbed droplets
        correctly.

        Args:
            other (SphericalDroplet): instance of the other droplet
            grid (GridBase): grid that determines how distances are measured,
                which is for instance important to respect periodic boundary
                conditions. If omitted, an Eucledian distance is assumed.

        Returns:
            bool: whether the droplets overlap or not
        """
        if grid is None:
            distance = np.linalg.norm(self.position - other.position)
        else:
            distance = grid.distance_real(self.position, other.position)
        return distance < self.radius + other.radius  # type: ignore

    @preserve_scalars
    def interface_position(self, φ):
        """calculates the position of the interface of the droplet

        Args:
            φ (float or array): The angle in the polar coordinate system that
                is used to describe the interface

        Returns:
            An array with the coordinates of the interfacial points associated
            with each angle given by `φ`.

        Raises:
            ValueError: If the dimension of the space is not 2
        """
        if self.dim != 2:
            raise ValueError("Interfacial position only supported for 2d grids")
        pos = self.radius * np.transpose([np.cos(φ), np.sin(φ)])
        return self.position[None, :] + pos

    @property
    def interface_curvature(self):
        """float: the mean curvature of the interface of the droplet"""
        return 1 / self.radius

    def _get_phase_field(self, grid: GridBase, dtype=np.double):
        """Creates an image of the droplet on the `grid`

        Args:
            grid (:class:`~pde.grids.base.GridBase`):
                The grid used for discretizing the droplet phase field
            dtype:
                The numpy data type defining the type of the returned image.
                If `dtype == np.bool`, a binary representation is returned.

        Returns:
            An array with data values representing the droplet phase field at
            support points of the `grid`.
        """
        if self.dim != grid.dim:
            raise ValueError(
                f"Droplet (dimension {self.dim}) incompatible with grid (dimension "
                f"{grid.dim})"
            )

        # calculate distances from droplet center
        dist = grid.polar_coordinates_real(self.position)
        return (dist < self.radius).astype(dtype)

    def get_phase_field(
        self, grid: GridBase, label: str = None, dtype=np.double
    ) -> ScalarField:
        """Creates an image of the droplet on the `grid`

        Args:
            grid (:class:`~pde.grids.base.GridBase`):
                The grid used for discretizing the droplet phase field
            label (str):
                The label associated with the returned scalar field
            dtype:
                The numpy data type of the returned array. Typical choices are
                `numpy.double` and `numpy.bool`.

        Returns:
            :class:`~pde.fields.ScalarField`: A scalar field
            representing the droplet
        """
        data = self._get_phase_field(grid, dtype=dtype)
        return ScalarField(grid, data=data, label=label)

    def get_phase_field_masked(self, grid: GridBase, mask):
        """Returns an image of the droplet restricted to the given `mask`

        Args:
            grid (:class:`~pde.grids.base.GridBase`):
                The grid used for discretizing the droplet phase field
            mask (array):
                A binary mask marking the support points whose data is returned

        Returns:
            An array with the values at the support points marked in `mask`
        """
        return self._get_phase_field(grid)[mask]

    def _get_mpl_patch(self, **kwargs):
        """ return the patch representing the droplet for plotting """
        import matplotlib as mpl

        if self.dim != 2:
            raise NotImplementedError("Plotting is only implemented in 2d")

        return mpl.patches.Circle(self.position, self.radius, **kwargs)

    @plot_on_axes()
    def plot(self, ax, **kwargs):
        """Plot the droplet

        Args:
            ax (:class:`matplotlib.axes.Axes`):
                The axes to which the droplet picture is added
            **kwargs:
                Additional keyword arguments are passed to the
                :class:`matplotlib.patches.Circle`, which creates the patch that
                represents the droplet
        """
        kwargs.setdefault("fill", False)
        ax.add_artist(self._get_mpl_patch(**kwargs))


class DiffuseDroplet(SphericalDroplet):
    """ Represents a single, spherical droplet with a diffuse interface """

    __slots__ = ["data"]

    def __init__(
        self, position: Sequence[float], radius: float, interface_width: float = None
    ):
        """
        Args:
            position (:class:`numpy.ndarray`):
                Position of the droplet center
            radius (float):
                Radius of the droplet
            interface_width (float, optional):
                Width of the interface
        """
        self._init_data(position=position)
        super().__init__(position=position, radius=radius)
        self.interface_width = interface_width

    @property
    def data_bounds(self):
        """ lower and upper bounds on the parameters """
        l, h = super().data_bounds
        l[self.dim + 1] = 0  # interface width must be non-negative
        return l, h

    @classmethod
    def get_dtype(cls, position):
        dtype = super().get_dtype(position=position)
        return dtype + [("interface_width", float)]

    @property
    def interface_width(self) -> Optional[float]:
        if np.isnan(self.data["interface_width"]):
            return None
        else:
            return float(self.data["interface_width"])

    @interface_width.setter
    def interface_width(self, value: Optional[float]):
        if value is None:
            self.data["interface_width"] = np.nan
        elif value < 0:
            raise ValueError("Interface width must not be negative")
        else:
            self.data["interface_width"] = value
        self.check_data()

    def _get_phase_field(self, grid: GridBase, dtype=np.double):
        """Creates an image of the droplet on the `grid`

        Args:
            grid (:class:`~pde.grids.base.GridBase`):
                The grid used for discretizing the droplet phase field
            dtype:
                The numpy data type defining the type of the returned image.
                If `dtype == np.bool`, a binary representation is returned.

        Returns:
            An array with data values representing the droplet phase field at
            support points of the `grid`.
        """
        if self.dim != grid.dim:
            raise ValueError(
                f"Droplet (dimension {self.dim}) incompatible with grid (dimension "
                f"{grid.dim})"
            )

        if self.interface_width is None:
            interface_width = grid.typical_discretization
        else:
            interface_width = self.interface_width

        # calculate distances from droplet center
        dist = grid.polar_coordinates_real(self.position)

        # make the image
        if interface_width == 0 or dtype == np.bool:
            result = dist < self.radius
        else:
            result = 0.5 + 0.5 * np.tanh((self.radius - dist) / interface_width)

        return result.astype(dtype)


class PerturbedDropletBase(DiffuseDroplet, metaclass=ABCMeta):
    """represents a single droplet with a perturbed shape.

    This acts as an abstract class for which member functions need to specified
    depending on dimensionality.
    """

    __slots__ = ["data"]

    def __init__(
        self,
        position: List[float],
        radius: float,
        interface_width: float = None,
        amplitudes: List[float] = None,
    ):
        """
        Args:
            position (:class:`numpy.ndarray`):
                Position of the droplet center
            radius (float):
                Radius of the droplet
            interface_width (float, optional):
                Width of the interface
            amplitudes (:class:`numpy.ndarray`):
                The amplitudes of the perturbations
        """
        self._init_data(position=position, amplitudes=amplitudes)
        super().__init__(
            position=position, radius=radius, interface_width=interface_width
        )

        self.amplitudes = amplitudes

        if len(self.position) != self.__class__.dim:
            raise ValueError(f"Space dimension must be {self.__class__.dim}")

    @classmethod
    def get_dtype(cls, position, amplitudes):
        dtype = super().get_dtype(position=position)

        if amplitudes is None:
            modes = 0
        else:
            modes = len(amplitudes)
        return dtype + [("amplitudes", float, (modes,))]

    @property
    def data_bounds(self):
        """ lower and upper bounds on the parameters """
        l, h = super().data_bounds
        n = self.dim + 2
        # relative perturbation amplitudes must be between [-1, 1]
        l[n : n + self.modes] = -1
        h[n : n + self.modes] = 1
        return l, h

    @property
    def modes(self) -> int:
        """ int: number of perturbation modes """
        shape = self.data.dtype.fields["amplitudes"][0].shape
        return int(shape[0]) if shape else 1

    @property
    def amplitudes(self):
        return np.atleast_1d(self.data["amplitudes"])

    @amplitudes.setter
    def amplitudes(self, value: float):
        self.data["amplitudes"] = np.broadcast_to(value, (self.modes,))
        self.check_data()

    @abstractmethod
    def interface_distance(self, *angles):
        pass

    @abstractmethod
    def interface_curvature(self, *angles):
        pass

    @property
    def volume(self) -> float:
        raise NotImplementedError

    @volume.setter
    def volume(self, volume: float):
        raise NotImplementedError

    @property
    def surface_area(self) -> float:
        raise NotImplementedError

    def _get_phase_field(self, grid: GridBase, dtype=np.double):
        """create an image of the droplet on the `grid`

        `dtype` defines the type of the returned image. If dtype == np.bool, a
            binary representation is returned.
        """
        if self.dim != grid.dim:
            raise ValueError(
                f"Droplet (dimension {self.dim}) incompatible with grid (dimension "
                f"{grid.dim})"
            )

        if self.interface_width is None:
            interface_width = grid.typical_discretization
        else:
            interface_width = self.interface_width

        # calculate grid distance from droplet center
        dist, *angles = grid.polar_coordinates_real(self.position, ret_angle=True)

        # calculate interface distance from droplet center
        interface = self.interface_distance(*angles)

        # make the image
        if interface_width == 0 or dtype == np.bool:
            result = dist < interface
        else:
            result = 0.5 + 0.5 * np.tanh((interface - dist) / interface_width)

        return result.astype(dtype)

    def get_phase_field(
        self, grid: GridBase, label: str = None, dtype=np.double
    ) -> ScalarField:
        """create an image of the droplet on the `grid`

        `dtype` defines the type of the returned image. If dtype == np.bool, a
            binary representation is returned.
        `label` sets an identifier for the phase field
        """
        data = self._get_phase_field(grid, dtype=dtype)
        return ScalarField(grid, data=data, label=label)

    def get_phase_field_masked(self, grid: GridBase, mask):
        """create an image of the droplet on the `grid` and return only the
        points that are non-zero in the mask"""
        return self._get_phase_field(grid)[mask]


class PerturbedDroplet2D(PerturbedDropletBase):
    r"""Represents a single droplet in two dimensions with a perturbed shape

    The shape is described using the distance :math:`R(\phi)` of the interface
    from the `position`, which is a function of the polar angle :math:`\phi`.
    This function is expressed as a truncated series of harmonics:

    .. math::
        R(\phi) = R_0 + R_0\sum_{n=1}^N \left[ \epsilon^{(1)}_n \sin(n \phi)
                                    + \epsilon^{(2)}_n \cos(n \phi) \right]

    where :math:`N` is the number of perturbation modes considered, which is
    given by half the length of the `amplitudes` array. Consequently, amplitudes
    should always be an even number, to consider both `sin` and `cos` terms.
    """

    dim = 2

    __slots__ = ["data"]

    def __init__(
        self,
        position: List[float],
        radius: float,
        interface_width: float = None,
        amplitudes: List[float] = None,
    ):
        r"""
        Args:
            position (:class:`numpy.ndarray`):
                Position of the droplet center
            radius (float):
                Radius of the droplet
            interface_width (float, optional):
                Width of the interface
            amplitudes (:class:`numpy.ndarray`):
                (dimensionless) perturbation amplitudes
                :math:`\{\epsilon^{(1)}_1, \epsilon^{(2)}_1, \epsilon^{(1)}_2,
                \epsilon^{(2)}_2, \epsilon^{(1)}_3, \epsilon^{(2)}_3, \dots \}`.
                The length of the array needs to be even to capture
                perturbations of the highest mode consistently.
        """
        super().__init__(position, radius, interface_width, amplitudes)
        if len(self.amplitudes) % 2 != 0:
            logger = logging.getLogger(self.__class__.__name__)
            logger.warning(
                "`amplitudes` should be of even length to capture all perturbations of "
                "the highest mode."
            )

    @preserve_scalars
    def interface_distance(self, φ):
        """calculates the distance of the droplet interface to the origin

        Args:
            φ (float or array): The angle in the polar coordinate system that
                is used to describe the interface

        Returns:
            An array with the distances of the interfacial points associated
            with each angle given by `φ`.
        """
        dist = np.ones(φ.shape, dtype=np.double)
        for n, (a, b) in enumerate(iterate_in_pairs(self.amplitudes), 1):
            if a != 0:
                dist += a * np.sin(n * φ)
            if b != 0:
                dist += b * np.cos(n * φ)
        return self.radius * dist

    @preserve_scalars
    def interface_position(self, φ):
        """calculates the position of the interface of the droplet

        Args:
            φ (float or array): The angle in the polar coordinate system that
                is used to describe the interface

        Returns:
            An array with the coordinates of the interfacial points associated
            with each angle given by `φ`.
        """
        dist = self.interface_distance(φ)
        pos = dist[:, None] * np.transpose([np.sin(φ), np.cos(φ)])
        return self.position[None, :] + pos

    @preserve_scalars
    def interface_curvature(self, φ):
        r"""calculates the mean curvature of the interface of the droplet

        For simplicity, the effect of the perturbations are only included to
        linear order in the perturbation amplitudes :math:`\epsilon^{(1/2)}_n`.

        Args:
            φ (float or array): The angle in the polar coordinate system that
                is used to describe the interface

        Returns:
            An array with the curvature at the interfacial points associated
            with each angle given by `φ`.
        """
        curv_radius = np.ones(φ.shape, dtype=np.double)
        for n, (a, b) in enumerate(iterate_in_pairs(self.amplitudes), 1):
            factor = n * n - 1
            if a != 0:
                curv_radius -= a * factor * np.sin(n * φ)
            if b != 0:
                curv_radius -= b * factor * np.cos(n * φ)
        return 1 / (self.radius * curv_radius)

    @property
    def volume(self) -> float:
        """ float: volume of the droplet """
        term = 1 + np.sum(self.amplitudes ** 2) / 2
        return np.pi * self.radius ** 2 * term  # type: ignore

    @volume.setter
    def volume(self, volume: float):
        """ set volume keeping relative perturbations """
        term = 1 + np.sum(self.amplitudes ** 2) / 2
        self.radius = np.sqrt(volume / (np.pi * term))

    @property
    def surface_area(self) -> float:
        """ float: surface area of the droplet """
        # discretize surface for simple approximation to integral
        φs, dφ = np.linspace(0, 2 * np.pi, 256, endpoint=False, retstep=True)

        dist = np.ones(φs.shape, dtype=np.double)
        dist_dφ = np.zeros(φs.shape, dtype=np.double)
        for n, (a, b) in enumerate(iterate_in_pairs(self.amplitudes), 1):
            if a != 0:
                dist += a * np.sin(n * φs)
                dist_dφ += a * n * np.cos(n * φs)
            if b != 0:
                dist += b * np.cos(n * φs)
                dist_dφ -= b * n * np.sin(n * φs)

        dx = dist_dφ * np.cos(φs) - dist * np.sin(φs)
        dy = dist_dφ * np.sin(φs) + dist * np.cos(φs)
        line_element = np.hypot(dx, dy)

        return self.radius * line_element.sum() * dφ  # type: ignore

    @property
    def surface_area_approx(self) -> float:
        """ float: surface area of the droplet (quadratic in amplitudes) """
        length = 4
        for n, (a, b) in enumerate(iterate_in_pairs(self.amplitudes), 1):
            length += n ** 2 * (a ** 2 + b ** 2)
        return np.pi * self.radius * length / 2  # type: ignore

    def _get_mpl_patch(self, **kwargs):
        """ return the patch representing the droplet for plotting """
        import matplotlib as mpl

        φ = np.linspace(0, 2 * np.pi, endpoint=False)
        xy = self.interface_position(φ)
        return mpl.patches.Polygon(xy, closed=True, **kwargs)

    @plot_on_axes()
    def plot(self, ax, **kwargs):
        """Plot the perturbed droplet

        Args:
            ax (:class:`matplotlib.axes.Axes`):
                The axes to which the droplet picture is added
            **kwargs:
                Additional keyword arguments are passed to the
                :class:`matplotlib.patches.Polygon`, which creates the patch that
                represents the droplet.
        """
        kwargs.setdefault("color", "k")
        ax.add_patch(self._get_mpl_patch(**kwargs))


class PerturbedDroplet3D(PerturbedDropletBase):
    r"""Represents a single droplet in two dimensions with a perturbed shape

    The shape is described using the distance :math:`R(\theta, \phi)` of the
    interface from the origin as a function of the azimuthal angle
    :math:`\theta` and the polar angle :math:`\phi`. This function is developed
    as a truncated series of spherical harmonics :math:`Y_{l,m}(\theta, \phi)`:

    .. math::
        R(\theta, \phi) = R_0 \left[1 + \sum_{l=1}^{N_l}\sum_{m=-l}^l
                                \epsilon_{l,m} Y_{l,m}(\theta, \phi) \right]

    where :math:`N_l` is the number of perturbation modes considered, which is
    deduced from the length of the `amplitudes` array.
    """

    dim = 3

    __slots__ = ["data"]

    def __init__(
        self,
        position: List[float],
        radius: float,
        interface_width: float = None,
        amplitudes: List[float] = None,
    ):
        r"""
        Args:
            position (:class:`numpy.ndarray`):
                Position of the droplet center
            radius (float):
                Radius of the droplet
            interface_width (float, optional):
                Width of the interface
            amplitudes (:class:`numpy.ndarray`):
                Perturbation amplitudes :math:`\epsilon_{l,m}`. The length of
                the array needs to be 0, 3, 8, 15, 24, ... to capture
                perturbations of the highest mode consistently.
        """
        super().__init__(position, radius, interface_width, amplitudes)
        num_modes = len(self.amplitudes) + 1
        if not spherical.spherical_index_count_optimal(num_modes):
            logger = logging.getLogger(self.__class__.__name__)
            l, _ = spherical.spherical_index_lm(num_modes)
            opt_modes = spherical.spherical_index_count(l) - 1
            logger.warning(
                "The length of `amplitudes` should be such that all orders are "
                f"captured for the perturbations with the highest degree ({l}). "
                f"Consider increasing the size of the array to {opt_modes}."
            )

    @preserve_scalars
    def interface_distance(self, θ, φ):
        r"""calculates the distance of the droplet interface to the origin

        Args:
            θ (float or array): Azimuthal angle (in :math:`[0, \pi]`)
            φ (float or array): Polar angle (in :math:`[0, 2\pi]`)

        Returns:
            An array with the distances of the interfacial points associated
            with the angles.
        """
        assert θ.shape == φ.shape
        dist = np.ones(φ.shape, dtype=np.double)
        for k, a in enumerate(self.amplitudes, 1):
            if a != 0:
                dist += a * spherical.spherical_harmonic_real_k(k, θ, φ)
        return self.radius * dist

    @preserve_scalars
    def interface_position(self, θ, φ):
        r"""calculates the position of the interface of the droplet

        Args:
            θ (float or array): Azimuthal angle (in :math:`[0, \pi]`)
            φ (float or array): Polar angle (in :math:`[0, 2\pi]`)

        Returns:
            An array with the coordinates of the interfacial points associated
            with the angles.
        """
        dist = self.interface_distance(θ, φ)
        unit_vector = [np.sin(θ) * np.cos(φ), np.sin(θ) * np.sin(φ), np.cos(θ)]
        pos = dist[:, None] * np.transpose(unit_vector)
        return self.position[None, :] + pos

    @preserve_scalars
    def interface_curvature(self, θ, φ):
        r"""calculates the mean curvature of the interface of the droplet

        For simplicity, the effect of the perturbations are only included to
        linear order in the perturbation amplitudes :math:`\epsilon_{l,m}`.

        Args:
            θ (float or array): Azimuthal angle (in :math:`[0, \pi]`)
            φ (float or array): Polar angle (in :math:`[0, 2\pi]`)

        Returns:
            An array with the curvature at the interfacial points associated
            with the angles
        """
        Yk = spherical.spherical_harmonic_real_k
        correction = 0
        for k, a in enumerate(self.amplitudes, 1):
            if a != 0:
                l, _ = spherical.spherical_index_lm(k)
                hk = (l ** 2 + l - 2) / 2
                correction = a * hk * Yk(k, θ, φ)
        return 1 / self.radius + correction / self.radius ** 2

    @property
    def volume(self) -> float:
        """ float: volume of the droplet (determined numerically) """

        def integrand(θ, φ):
            """ helper function calculating the integrand """
            r = self.interface_distance(θ, φ)
            return r ** 3 * np.sin(θ) / 3

        volume = integrate.dblquad(
            integrand, 0, 2 * np.pi, lambda _: 0, lambda _: np.pi
        )[0]
        return volume  # type: ignore

    @volume.setter
    def volume(self, volume: float):
        """ set volume keeping relative perturbations """
        raise NotImplementedError("Cannot set volume")

    @property
    def volume_approx(self) -> float:
        """ float: approximate volume to linear order in the perturbation """
        volume = spherical.volume_from_radius(self.radius, 3)
        if len(self.amplitudes) > 0:
            volume += self.amplitudes[0] * 2 * np.sqrt(np.pi) * self.radius ** 2
        return volume


def droplet_from_data(droplet_class: str, data) -> DropletBase:
    """create a droplet instance of the given class using some data

    Args:
        droplet_class (str): The name of the class that is used to create the
            droplet instance
        data (numpy.ndarray): A numpy array that defines the droplet properties
    """
    cls = DropletBase._subclasses[droplet_class]
    return cls(**{key: data[key] for key in data.dtype.names})  # type: ignore


__all__ = [
    "SphericalDroplet",
    "DiffuseDroplet",
    "PerturbedDroplet2D",
    "PerturbedDroplet3D",
]



"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import numpy as np
import pytest

from pde import ScalarField, UnitGrid
from pde.tools.misc import skipUnlessModule

from .. import DiffuseDroplet, Emulsion, SphericalDroplet, droplets, emulsions


def test_empty_emulsion():
    """ test an emulsions without any droplets """
    e = Emulsion([], grid=UnitGrid([2]))
    assert not e
    assert len(e) == 0
    assert e == e.copy()
    assert e is not e.copy()
    assert e.interface_width is None
    assert e.total_droplet_volume == 0
    dists = e.get_pairwise_distances()
    np.testing.assert_array_equal(dists, np.zeros((0, 0)))
    expect = {
        "count": 0,
        "radius_mean": np.nan,
        "radius_std": np.nan,
        "volume_mean": np.nan,
        "volume_std": np.nan,
    }
    assert e.get_size_statistics() == expect
    for b in [True, False]:
        np.testing.assert_array_equal(e.get_neighbor_distances(b), np.array([]))


def test_emulsion_single():
    """ test an emulsions with a single droplet """
    e = Emulsion([], grid=UnitGrid([2]))
    e.append(DiffuseDroplet([10], 3, 1))
    assert e
    assert len(e) == 1
    assert e == e.copy()
    assert e is not e.copy()
    assert e.interface_width == pytest.approx(1)
    assert e.total_droplet_volume == pytest.approx(6)
    dists = e.get_pairwise_distances()
    np.testing.assert_array_equal(dists, np.zeros((1, 1)))
    expect = {
        "count": 1,
        "radius_mean": 3,
        "radius_std": 0,
        "volume_mean": 6,
        "volume_std": 0,
    }
    assert e.get_size_statistics() == expect
    for b in [True, False]:
        np.testing.assert_array_equal(e.get_neighbor_distances(b), np.array([np.nan]))


def test_emulsion_two():
    """ test an emulsions with two droplets """
    grid = UnitGrid([30])
    e = Emulsion([DiffuseDroplet([10], 3, 1)], grid=grid)
    e1 = Emulsion([DiffuseDroplet([20], 5, 1)], grid=grid)
    e.extend(e1)
    assert e
    assert len(e) == 2
    assert e == e.copy()
    assert e is not e.copy()
    assert e.interface_width == pytest.approx(1)
    assert e.total_droplet_volume == pytest.approx(16)

    dists = e.get_pairwise_distances()
    np.testing.assert_array_equal(dists, np.array([[0, 10], [10, 0]]))
    expect = {
        "count": 2,
        "radius_mean": 4,
        "radius_std": 1,
        "volume_mean": 8,
        "volume_std": 2,
    }
    assert e.get_size_statistics() == expect

    np.testing.assert_array_equal(e.get_neighbor_distances(False), np.array([10, 10]))
    np.testing.assert_array_equal(e.get_neighbor_distances(True), np.array([2, 2]))


def test_emulsion_incompatible():
    """ test incompatible droplets in an emulsion """
    # different type
    d1 = SphericalDroplet([1], 2)
    d2 = DiffuseDroplet([1], 2, 1)
    e = Emulsion([d1, d2])
    assert len(e) == 2
    with pytest.raises(TypeError):
        e.data

    # same type
    d1 = SphericalDroplet([1], 2)
    d2 = SphericalDroplet([1, 2], 2)
    with pytest.raises(ValueError):
        e = Emulsion([d1, d2])


def test_emulsion_linked_data():
    """ test whether emulsions link the data to droplets correctly """
    d1 = SphericalDroplet([0, 0], 1)
    d2 = SphericalDroplet([1, 2], 3)
    e = Emulsion([d1, d2])
    data = e.get_linked_data()

    assert e[0] == d1
    assert e[1] == d2

    data[1] = 5
    assert e[0] == d1
    assert e[1] != d2
    np.testing.assert_array_equal(e[1]._data_array, 5)


@skipUnlessModule("h5py")
def test_emulsion_io(tmp_path):
    """ test writing and reading emulsions """
    path = tmp_path / "test_emulsion_io.hdf5"

    es = [
        Emulsion(),
        Emulsion([DiffuseDroplet([0, 1], 10, 0.5)] * 2),
        Emulsion([droplets.PerturbedDroplet2D([0, 1], 3, 1, [1, 2, 3])]),
    ]
    for e1 in es:
        e1.to_file(path)
        e2 = Emulsion.from_file(path)
        assert e1 == e2


def test_timecourse():
    """ test some droplet track functions """
    t1 = emulsions.EmulsionTimeCourse()
    for i in range(4):
        d = droplets.SphericalDroplet([i], i)
        t1.append([d], i)

    for track, length in [(t1, 4), (t1[:2], 2)]:
        assert len(track) == length
        assert track.times == list(range(length))
        assert [t for t, _ in track.items()] == list(range(length))

    assert t1 == t1[:]
    assert t1[3:3] == emulsions.EmulsionTimeCourse()

    t2 = emulsions.EmulsionTimeCourse(t1)
    assert t1 == t2
    assert t1 is not t2

    t1.clear()
    assert len(t1) == 0


@skipUnlessModule("h5py")
def test_timecourse_io(tmp_path):
    """ test writing and reading emulsions time courses """
    path = tmp_path / "test_timecourse_io.hdf5"

    e1 = Emulsion()
    e2 = Emulsion([DiffuseDroplet([0, 1], 10, 0.5)] * 2)
    tc1 = emulsions.EmulsionTimeCourse([e1, e2], times=[0, 10])

    tc1.to_file(path)
    tc2 = emulsions.EmulsionTimeCourse.from_file(path, progress=False)
    assert tc1.times == tc2.times
    assert tc1.emulsions == tc2.emulsions
    assert len(tc2) == 2


@skipUnlessModule("matplotlib")
def test_emulsion_plotting():
    """ test plotting emulsions """
    field = ScalarField(UnitGrid([10, 10], periodic=True))
    es = [
        Emulsion([DiffuseDroplet([0, 1], 10, 0.5)] * 2),
        Emulsion([droplets.PerturbedDroplet2D([0, 1], 3, 1, [1, 2, 3])]),
    ]
    for e1 in es:
        e1.plot()
        e1.plot(field=field)

    with pytest.raises(NotImplementedError):
        Emulsion().plot()


def test_remove_overlapping():
    """ test that removing overlapping droplets works """
    e = Emulsion([SphericalDroplet([0, 1], 2), SphericalDroplet([1, 1], 2)])
    assert len(e) == 2
    e.remove_overlapping()
    assert len(e) == 1



"""
Functions and classes for analyzing emulsions and droplets

.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

from .droplet_tracks import DropletTrack, DropletTrackList
from .droplets import DiffuseDroplet, SphericalDroplet
from .emulsions import Emulsion, EmulsionTimeCourse
from .image_analysis import get_length_scale, get_structure_factor, locate_droplets
from .trackers import DropletTracker, LengthScaleTracker
from .version import __version__



"""
.. codeauthor:: David Zwicker <david.zwicker@ds.mpg.de>
"""

import numpy as np
import pytest

from pde import CahnHilliardPDE, CartesianGrid, DiffusionPDE, ScalarField, UnitGrid
from pde.tools.misc import skipUnlessModule

from .. import LengthScaleTracker, SphericalDroplet
from ..emulsions import EmulsionTimeCourse


@skipUnlessModule("h5py")
def test_emulsion_tracker(tmp_path):
    """ test using the emulsions tracker """
    path = tmp_path / "test_emulsion_tracker.hdf5"

    d = SphericalDroplet([4, 4], 3)
    c = d.get_phase_field(UnitGrid([8, 8]))

    pde = CahnHilliardPDE()

    e1 = EmulsionTimeCourse()
    tracker = e1.tracker(filename=path)
    pde.solve(c, t_range=1, dt=1e-3, backend="numpy", tracker=tracker)
    e2 = EmulsionTimeCourse.from_file(path, progress=False)

    assert e1 == e2
    assert len(e1) == 2
    assert len(e1[0]) == 1  # found a single droplet
    assert path.stat().st_size > 0  # wrote some result


def test_length_scale_tracker(tmp_path):
    """ test the length scale tracker """
    path = tmp_path / "test_length_scale_tracker.json"

    grid = CartesianGrid([[0, 10 * np.pi]], 64, periodic=True)
    field = ScalarField.from_expression(grid, "sin(2 * x)")

    pde = DiffusionPDE()
    tracker = LengthScaleTracker(0.05, filename=path)
    pde.solve(field, t_range=0.1, backend="numpy", tracker=tracker)

    for ls in tracker.length_scales:
        assert ls == pytest.approx(np.pi, rel=1e-3)
    assert path.stat().st_size > 0  # wrote some result



from pathlib import Path

from setuptools import find_packages, setup

BASE_PATH = Path(__file__).resolve().parent


# read the version from the particular file
with open(BASE_PATH / "droplets" / "version.py", "r") as f:
    exec(f.read())

DOWNLOAD_URL = (
    f"https://github.com/zwicker-group/py-droplets/archive/v{__version__}.tar.gz"
)


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="py-droplets",
    package_data={"droplets": ["py.typed"]},
    packages=find_packages(),
    zip_safe=False,  # this is required for mypy to find the py.typed file
    version=__version__,
    license="MIT",
    description=(
        "Python package for describing and analyzing droplets in experiments and "
        "simulations"
    ),
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="David Zwicker",
    author_email="david.zwicker@ds.mpg.de",
    url="https://github.com/zwicker-group/py-droplets",
    download_url=DOWNLOAD_URL,
    keywords=["emulsions", "image-analysis"],
    python_requires=">=3.6",
    install_requires=["matplotlib", "numpy", "numba", "scipy", "sympy", "py-pde"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)
